
import os
from PyQt4.QtCore import *
from PyQt4.QtGui import *

from sfa.util.rspecHelper import RSpec
from sface.config import config
from sface.sfiprocess import SfiProcess
from sface.screens.sfascreen import SfaScreen

already_in_nodes = []

node_status = { "in": "Already Selected",
                "out": "Not Selected",
                "add": "To be Added",
                "remove": "To be Removed"}

tag_status = { "in": "Already Set",
                "out": "Not Set",
                "add": "To be Added",
                "remove": "To be Removed"}

default_tags = "Default tags"
settable_tags = ['delegations', 'initscript']

def itemType(index):
    if index.parent().parent().isValid():
        return "tag"
    else:
        return "node"


class NodeView(QTreeView):
    def __init__(self, parent):
        QTreeView.__init__(self, parent)

        self.setAnimated(True)
        self.setItemsExpandable(True)
        self.setRootIsDecorated(True)
        self.setAlternatingRowColors(True)
#        self.setSelectionMode(self.MultiSelection)
        self.setAttribute(Qt.WA_MacShowFocusRect, 0)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.setToolTip("Double click on a row to change its status.  Right click on a host to add a tag.")

    def mouseDoubleClickEvent(self, event):
        index = self.currentIndex()
        model = index.model()
        status_index = model.index(index.row(), 1, index.parent())
        status_data = status_index.data().toString()
        node_index = model.index(index.row(), 0, index.parent())
        node_data = node_index.data().toString()

        if itemType(node_index) == "tag":
            data = node_index.data().toString()
            tagname, value = data.split(": ")
            if tagname not in settable_tags:
                # Pop up error msg
                QMessageBox.warning(self, "Not settable", "Insufficient permission to change '%s' tag" % tagname)
                return
            if status_data == tag_status['in']:
                model.setData(status_index, QString(tag_status['remove']))
            elif status_data == tag_status['add']:
                model.setData(status_index, QString(tag_status['out']))
            elif status_data == tag_status['remove']:
                model.setData(status_index, QString(tag_status['in']))
            else: model.setData(status_index, QString(node_status['out']))
        else:
            # This is a hostname
            if status_data == node_status['in']:
                model.setData(status_index, QString(node_status['remove']))
            elif status_data == node_status['out']:
                model.setData(status_index, QString(node_status['add']))
            elif status_data in (node_status['add'], node_status['remove']):
                if node_data in already_in_nodes: model.setData(status_index, QString(node_status['in']))
                else: model.setData(status_index, QString(node_status['out']))

        model.emit(SIGNAL("dataChanged(QModelIndex, QModelIndex)"), node_index, node_index)

    def mousePressEvent(self, event):
        QTreeView.mousePressEvent(self, event)
        if event.button() == Qt.LeftButton:
            return

        # Right click
        index = self.currentIndex()
        model = index.model()
        status_index = model.index(index.row(), 1, index.parent())
        status_data = status_index.data().toString()
        node_index = model.index(index.row(), 0, index.parent())
        node_data = node_index.data().toString()

        if itemType(node_index) == "node":
            # This is a hostname
            if status_data in (node_status['in'], node_status['add'], ""):
                # Pop up a dialog box for adding a new attribute
                tagname, ok = QInputDialog.getItem(self, "Add tag",
                                                   "Tag name:", settable_tags)
                if ok:
                    value, ok = QInputDialog.getText(self, "Add tag",
                                                     "Value for tag '%s'" % tagname)
                    if ok:
                        # Add a new row to the model for the tag

                        # For testing with the QStandardItemModel
                        #nodeItem = model.itemFromIndex(index)
                        #tagstring = QString("%s: %s" % (tagname, value))
                        #tagItem = QStandardItem(tagstring)
                        #status = QStandardItem(QString(tag_status['add']))
                        #nodeItem.appendRow([tagItem, status])

                        # We're using the QSortFilterProxyModel here
                        src_index = model.mapToSource(index)
                        src_model = src_index.model()
                        nodeItem = src_model.itemFromIndex(src_index)
                        tagstring = QString("%s: %s" % (tagname, value))
                        tagItem = QStandardItem(tagstring)
                        status = QStandardItem(QString(tag_status['add']))
                        nodeItem.appendRow([tagItem, status])

            elif status_data in (node_status['out'], node_status['remove']):
                QMessageBox.warning(self, "Not selected", "Can only add tags to selected nodes")
                return

        model.emit(SIGNAL("dataChanged(QModelIndex, QModelIndex)"), node_index, node_index)

    def currentChanged(self, current, previous):
        model = current.model()
        node_index = model.index(current.row(), 0, current.parent())
        node_data = node_index.data().toString()
        self.emit(SIGNAL('hostnameClicked(QString)'), node_data)
        
                

class NodeNameDelegate(QStyledItemDelegate):
    def __init__(self, parent):
        QStyledItemDelegate.__init__(self, parent)

    def paint(self, painter, option, index):
        model = index.model()
        status_index = model.index(index.row(), 1, index.parent())
        status_data = status_index.data().toString()

        fm = QFontMetrics(option.font)
        rect = option.rect

        data = index.data().toString()
        rect.setHeight(rect.height() - 2)
        rect.setWidth(fm.width(QString(data)) + 6)
        rect.setX(rect.x() + 5)
        rect.setY(rect.y() - 1)

        x, y, h, w = rect.x(), rect.y(), rect.height(), rect.width()

        path = QPainterPath()
        path.addRoundedRect(x - 1, y + 1, w, h, 4, 4)

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)

        if itemType(index) == "node":
            if status_data == node_status['in']: # already in the slice
                painter.fillPath(path, QColor.fromRgb(0, 250, 250))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            elif status_data == node_status['add']: # newly added to the slice
                painter.fillPath(path, QColor.fromRgb(0, 250, 0))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            elif status_data == node_status['remove']: # removed from the slice
                painter.fillPath(path, QColor.fromRgb(250, 0, 0))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            else:
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

        else:
            if status_data == tag_status['in']: # already in the slice
                painter.fillPath(path, QColor.fromRgb(0, 250, 250))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            elif status_data == tag_status['add']: # newly added to the slice
                painter.fillPath(path, QColor.fromRgb(0, 250, 0))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            elif status_data == tag_status['remove']: # removed from the slice
                painter.fillPath(path, QColor.fromRgb(250, 0, 0))
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

            else:
                painter.setPen(QColor.fromRgb(0, 0, 0))
                painter.drawText(option.rect, 0, QString(data))

        painter.restore()

class SliceWidget(QWidget):
    def __init__(self, parent):
        QWidget.__init__(self, parent)

        self.network_names = []
        self.process = SfiProcess(self)

        self.slicename = QLabel("", self)
        self.updateSliceName()
        self.slicename.setScaledContents(False)
        searchlabel = QLabel ("Search: ", self)
        searchlabel.setScaledContents(False)
        searchbox = QLineEdit(self)
        searchbox.setAttribute(Qt.WA_MacShowFocusRect, 0)

        toplayout = QHBoxLayout()
        toplayout.addWidget(self.slicename, 0, Qt.AlignLeft)
        toplayout.addStretch()
        toplayout.addWidget(searchlabel, 0, Qt.AlignRight)
        toplayout.addWidget(searchbox, 0, Qt.AlignRight)

        self.nodeView = NodeView(self)
        self.nodeModel = QStandardItemModel(0, 2, self)
        self.filterModel = QSortFilterProxyModel(self) # enable filtering

        self.nodeNameDelegate = NodeNameDelegate(self)

        refresh = QPushButton("Update Slice Data", self)
        refresh.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)
        submit = QPushButton("Submit", self)
        submit.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Maximum)

        bottomlayout = QHBoxLayout()
        bottomlayout.addWidget(refresh, 0, Qt.AlignLeft)
        bottomlayout.addStretch()
        bottomlayout.addWidget(submit, 0, Qt.AlignRight)

        layout = QVBoxLayout()
        layout.addLayout(toplayout)
        layout.addWidget(self.nodeView)
        layout.addLayout(bottomlayout)
        self.setLayout(layout)
        self.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)

        self.connect(refresh, SIGNAL('clicked()'), self.refresh)
        self.connect(submit, SIGNAL('clicked()'), self.submit)
        self.connect(searchbox, SIGNAL('textChanged(QString)'), self.filter)
        self.connect(self.nodeView, SIGNAL('hostnameClicked(QString)'),
                     self.nodeSelectionChanged)

        self.updateView()

    def submitFinished(self):
        self.setStatus("<font color='green'>Slice data submitted.</font>")
        QTimer.singleShot(1000, self.refresh)

    def refreshFinished(self):
        self.setStatus("<font color='green'>Slice data updated.</font>", timeout=5000)
        self.updateView()
        self.parent().signalAll("rspecUpdated")

    def readSliceRSpec(self):
        rspec_file = config.getSliceRSpecFile()
        if os.path.exists(rspec_file):
            xml = open(rspec_file).read()
            return RSpec(xml)
        return None

    def setStatus(self, msg, timeout=None):
        self.parent().setStatus(msg, timeout)

    def checkRunningProcess(self):
        if self.process.isRunning():
            self.setStatus("<font color='red'>There is already a process running. Please wait.</font>")
            return True
        return False

    def filter(self, filter_string):
        # for hierarchical models QSortFilterProxyModel applies the
        # sort recursively. if the parent doesn't match the criteria
        # we won't be able to match the children. so we need to match
        # parent (by matching the network_names)
        networks = ["^%s$" % n for n in self.network_names]
        filters = networks + [str(filter_string)]
        self.filterModel.setFilterRegExp(QRegExp('|'.join(filters)))

    def itemStatus(self, item):
        statusItem = item.parent().child(item.row(), 1)
        return statusItem.data(Qt.DisplayRole).toString()

    def itemText(self, item):
        return item.data(Qt.DisplayRole).toString()

    # Recursively walk the tree, making changes to the RSpec
    def process_subtree(self, rspec, item, depth = 0):
        change = False
        model = self.nodeModel

        if depth in [0, 1]:
            pass
        elif depth == 2: # Hostname
            hostname = self.itemText(item)
            testbed = self.itemText(item.parent())
            status = self.itemStatus(item)
            if status == node_status['add']:
                print "Add hostname: %s" % hostname
                rspec.add_sliver(hostname, testbed)
                change = True
            elif status == node_status['remove']:
                print "Remove hostname: %s" % hostname
                rspec.remove_sliver(hostname, testbed)
                change = True
        elif depth == 3: # Tag
            tag, value = self.itemText(item).split(": ")
            status = self.itemStatus(item)
            tag = "%s" % tag     # Prevent weird error from lxml
            value = "%s" % value # Prevent weird error from lxml
            node = self.itemText(item.parent())
            testbed = self.itemText(item.parent().parent())
            if status == tag_status['add']:
                print "Add tag to (%s, %s): %s/%s " % (testbed, node, tag, value)
                if node.startsWith(default_tags):
                    rspec.add_default_sliver_attribute(tag, value, testbed)
                else:
                    rspec.add_sliver_attribute(node, tag, value, testbed)
                change = True
            elif status == tag_status['remove']:
                print "Remove tag from (%s, %s): %s/%s " % (testbed, node, tag, value)
                if node.startsWith(default_tags):
                    rspec.remove_default_sliver_attribute(tag, value, testbed)
                else:
                    rspec.remove_sliver_attribute(node, tag, value, testbed)
                change = True

        children = item.rowCount()
        for row in range(0, children):
            status = self.process_subtree(rspec, item.child(row), depth + 1)
            change = change or status

        return change

    def submit(self):
        if self.checkRunningProcess():
            return

        rspec = self.readSliceRSpec()
        change = self.process_subtree(rspec, self.nodeModel.invisibleRootItem())

        if not change:
            self.setStatus("<font color=red>No change in slice data. Not submitting!</font>", timeout=3000)
            return

        self.disconnect(self.process, SIGNAL('finished()'), self.refreshFinished)
        self.connect(self.process, SIGNAL('finished()'), self.submitFinished)

        self.process.applyRSpec(rspec)
        self.setStatus("Sending slice data (RSpec). This will take some time...")
        

    def refresh(self):
        if not config.getSlice():
            self.setStatus("<font color='red'>Slice not set yet!</font>")
            return

        if self.process.isRunning():
            self.setStatus("<font color='red'>There is already a process running. Please wait.</font>")
            return

        self.disconnect(self.process, SIGNAL('finished()'), self.submitFinished)
        self.connect(self.process, SIGNAL('finished()'), self.refreshFinished)

        self.process.getRSpecFromSM()
        self.setStatus("Updating slice data. This will take some time...")

    def updateView(self):
        global already_in_nodes
        already_in_nodes = []
        self.network_names = []
        self.nodeModel.clear()

        rspec = self.readSliceRSpec()
        if not rspec:
            return None

        rootItem = self.nodeModel.invisibleRootItem()
        networks = sorted(rspec.get_network_list())
        for network in networks:
            self.network_names.append(network)

            all_nodes = rspec.get_node_list(network)
            sliver_nodes = rspec.get_sliver_list(network)
            available_nodes = [ node for node in all_nodes if node not in sliver_nodes ]

            networkItem = QStandardItem(QString(network))
            msg = "%s Nodes\t%s Selected" % (len(all_nodes), len(sliver_nodes))
            rootItem.appendRow([networkItem, QStandardItem(QString(msg))])

            already_in_nodes += sliver_nodes

            # Add default slice tags
            nodeItem = QStandardItem(QString("%s for %s" % (default_tags, network)))
            statusItem = QStandardItem(QString(""))
            networkItem.appendRow([nodeItem, statusItem])
            attrs = rspec.get_default_sliver_attributes(network)
            for (name, value) in attrs:
                    tagstring = QString("%s: %s" % (name, value))
                    tagItem = QStandardItem(tagstring)
                    status = QStandardItem(QString(tag_status['in']))
                    nodeItem.appendRow([tagItem, status])

            for node in sliver_nodes:
                nodeItem = QStandardItem(QString(node))
                statusItem = QStandardItem(QString(node_status['in']))
                networkItem.appendRow([nodeItem, statusItem])

                attrs = rspec.get_sliver_attributes(node, network)
                for (name, value) in attrs:
                    tagstring = QString("%s: %s" % (name, value))
                    tagItem = QStandardItem(tagstring)
                    statusItem = QStandardItem(QString(tag_status['in']))
                    nodeItem.appendRow([tagItem, statusItem])

            for node in available_nodes:
                nodeItem = QStandardItem(QString(node))
                statusItem = QStandardItem(QString(node_status['out']))
                networkItem.appendRow([nodeItem, statusItem])

        self.filterModel.setSourceModel(self.nodeModel)
        self.filterModel.setFilterKeyColumn(-1)
        self.filterModel.setDynamicSortFilter(True)

        headers = QStringList() << "Hostname or Tag" << "Status"
        self.nodeModel.setHorizontalHeaderLabels(headers)

        self.nodeView.setItemDelegateForColumn(0, self.nodeNameDelegate)
        self.nodeView.setModel(self.filterModel)
        self.nodeView.expandAll()
        self.nodeView.resizeColumnToContents(0)
        self.nodeView.collapseAll()

    def updateSliceName(self):
        self.slicename.setText("Slice : %s" % (config.getSlice() or "None"))

    def nodeSelectionChanged(self, hostname):
        self.parent().nodeSelectionChanged(hostname)

class MainScreen(SfaScreen):
    def __init__(self, parent):
        SfaScreen.__init__(self, parent)

        slice = SliceWidget(self)
        self.init(slice, "Main Window", "OneLab SFA crawler")

    def rspecUpdated(self):
        self.mainwin.rspecWindow.updateView()
        
    def configurationChanged(self):
        self.widget.updateSliceName()
        self.widget.updateView()
        self.mainwin.rspecWindow.updateView()

    def nodeSelectionChanged(self, hostname):
        self.mainwin.nodeSelectionChanged(hostname)
